compctl -k "(
--apply \
--argument_groups \
--async_read \
--attribs \
--chunksize \
--chunkspec \
--cmor \
--config \
--copy_chunkspec \
--disable_file_suffix \
--disable_history \
--double \
--download_path \
--eccodes \
--enableexcept \
--envvars \
--file_suffix \
--force \
--gridsearchradius \
--history \
--history_info \
--icon_grids \
--ignore_time_bounds \
--module_info \
--netcdf_hdr_pad \
--no_history \
--no_remap_weights \
--nofile \
--operators \
--operators_no_output \
--pedantic \
--percentile \
--precision \
--print_filename \
--query \
--reduce_dim \
--remap_weights \
--remove_chunkspec \
--reset_history \
--rusage \
--seed \
--settings \
--shuffle \
--single \
--sortparam \
--test \
--timestat_date \
--use_fftw \
--use_time_bounds \
--version_info \
--worker \
-A, \
-C, \
-D, \
-F, \
-M, \
-O, \
-P, \
-Q, \
-R, \
-S, \
-T, \
-V, \
-Z, \
-a, \
-b, \
-c, \
-d, \
-f, \
-g, \
-h, \
-i, \
-k, \
-l, \
-m, \
-p, \
-r, \
-s, \
-t, \
-u, \
-v, \
-w, \
-z, \
abs -abs \
acos -acos \
add -add \
addc -addc \
addtrend -addtrend \
adipot -adipot \
adisit -adisit \
aexpr -aexpr \
aexprf -aexprf \
after -after \
afterburner -afterburner \
air_density -air_density \
anomaly -anomaly \
ap2hl -ap2hl \
ap2hlx -ap2hlx \
ap2pl -ap2pl \
ap2plx -ap2plx \
arg -arg \
asin -asin \
atan -atan \
atan2 -atan2 \
bandpass -bandpass \
bitrounding -bitrounding \
bottomvalue -bottomvalue \
boxavg -boxavg \
cat -cat \
cdiread -cdiread \
cdiwrite -cdiwrite \
changemulti -changemulti \
chcode -chcode \
chlevel -chlevel \
chlevelc -chlevelc \
chlevelv -chlevelv \
chltype -chltype \
chname -chname \
chparam -chparam \
chtabnum -chtabnum \
chunit -chunit \
chvar -chvar \
cinfo -cinfo \
clone -clone \
cloudlayer -cloudlayer \
cmor -cmor \
cmorlite -cmorlite \
codetab -codetab \
collgrid -collgrid \
complextopol -complextopol \
complextorect -complextorect \
conj -conj \
consecsum -consecsum \
consects -consects \
const -const \
contour -contour \
conv_cmor_table -conv_cmor_table \
copy -copy \
cos -cos \
coshill -coshill \
dayadd -dayadd \
dayavg -dayavg \
daycount -daycount \
daydiv -daydiv \
daymax -daymax \
daymean -daymean \
daymin -daymin \
daymul -daymul \
daypctl -daypctl \
dayrange -dayrange \
daystd -daystd \
daystd1 -daystd1 \
daysub -daysub \
daysum -daysum \
dayvar -dayvar \
dayvar1 -dayvar1 \
dcw -dcw \
del29feb -del29feb \
delattribute -delattribute \
delcode -delcode \
delday -delday \
delete -delete \
delgridcell -delgridcell \
delmulti -delmulti \
delname -delname \
delparam -delparam \
delta_pressure -delta_pressure \
deltat -deltat \
delvar -delvar \
detrend -detrend \
dhouravg -dhouravg \
dhourmax -dhourmax \
dhourmean -dhourmean \
dhourmin -dhourmin \
dhourrange -dhourrange \
dhourstd -dhourstd \
dhourstd1 -dhourstd1 \
dhoursum -dhoursum \
dhourvar -dhourvar \
dhourvar1 -dhourvar1 \
diff -diff \
diffc -diffc \
diffn -diffn \
diffp -diffp \
difftest -difftest \
diffv -diffv \
distgrid -distgrid \
div -div \
divc -divc \
divcoslat -divcoslat \
divdpm -divdpm \
divdpy -divdpy \
dminuteavg -dminuteavg \
dminutemax -dminutemax \
dminutemean -dminutemean \
dminutemin -dminutemin \
dminuterange -dminuterange \
dminutestd -dminutestd \
dminutestd1 -dminutestd1 \
dminutesum -dminutesum \
dminutevar -dminutevar \
dminutevar1 -dminutevar1 \
dump_cmor_table -dump_cmor_table \
dumpmap -dumpmap \
duplicate -duplicate \
dv2ps -dv2ps \
dv2uv -dv2uv \
dv2uvl -dv2uvl \
eca_cdd -eca_cdd \
eca_cfd -eca_cfd \
eca_csu -eca_csu \
eca_cwd -eca_cwd \
eca_cwdi -eca_cwdi \
eca_cwfi -eca_cwfi \
eca_etr -eca_etr \
eca_fd -eca_fd \
eca_gsl -eca_gsl \
eca_hd -eca_hd \
eca_hwdi -eca_hwdi \
eca_hwfi -eca_hwfi \
eca_id -eca_id \
eca_pd -eca_pd \
eca_r10mm -eca_r10mm \
eca_r1mm -eca_r1mm \
eca_r20mm -eca_r20mm \
eca_r75p -eca_r75p \
eca_r75ptot -eca_r75ptot \
eca_r90p -eca_r90p \
eca_r90ptot -eca_r90ptot \
eca_r95p -eca_r95p \
eca_r95ptot -eca_r95ptot \
eca_r99p -eca_r99p \
eca_r99ptot -eca_r99ptot \
eca_rr1 -eca_rr1 \
eca_rx1day -eca_rx1day \
eca_rx5day -eca_rx5day \
eca_sdii -eca_sdii \
eca_su -eca_su \
eca_tg10p -eca_tg10p \
eca_tg90p -eca_tg90p \
eca_tn10p -eca_tn10p \
eca_tn90p -eca_tn90p \
eca_tr -eca_tr \
eca_tx10p -eca_tx10p \
eca_tx90p -eca_tx90p \
enlarge -enlarge \
ensavg -ensavg \
ensbrs -ensbrs \
enscrps -enscrps \
enskurt -enskurt \
ensmax -ensmax \
ensmean -ensmean \
ensmedian -ensmedian \
ensmin -ensmin \
enspctl -enspctl \
ensrange -ensrange \
ensrkhistspace -ensrkhistspace \
ensrkhisttime -ensrkhisttime \
ensroc -ensroc \
ensskew -ensskew \
ensstd -ensstd \
ensstd1 -ensstd1 \
enssum -enssum \
ensvar -ensvar \
ensvar1 -ensvar1 \
eof -eof \
eof3d -eof3d \
eof3dspatial -eof3dspatial \
eof3dtime -eof3dtime \
eofcoeff -eofcoeff \
eofcoeff3d -eofcoeff3d \
eofspatial -eofspatial \
eoftime -eoftime \
eq -eq \
eqc -eqc \
etccdi -etccdi \
etccdi_cdd -etccdi_cdd \
etccdi_csdi -etccdi_csdi \
etccdi_cwd -etccdi_cwd \
etccdi_fd -etccdi_fd \
etccdi_gsl -etccdi_gsl \
etccdi_hd -etccdi_hd \
etccdi_id -etccdi_id \
etccdi_r10mm -etccdi_r10mm \
etccdi_r1mm -etccdi_r1mm \
etccdi_r20mm -etccdi_r20mm \
etccdi_r95p -etccdi_r95p \
etccdi_r99p -etccdi_r99p \
etccdi_rx1day -etccdi_rx1day \
etccdi_rx1daymon -etccdi_rx1daymon \
etccdi_rx5day -etccdi_rx5day \
etccdi_rx5daymon -etccdi_rx5daymon \
etccdi_sdii -etccdi_sdii \
etccdi_su -etccdi_su \
etccdi_tn10p -etccdi_tn10p \
etccdi_tn90p -etccdi_tn90p \
etccdi_tr -etccdi_tr \
etccdi_tx10p -etccdi_tx10p \
etccdi_tx90p -etccdi_tx90p \
etccdi_wsdi -etccdi_wsdi \
exp -exp \
export_e5ml -export_e5ml \
expr -expr \
exprf -exprf \
fc2gp -fc2gp \
fc2sp -fc2sp \
fdns -fdns \
filedes -filedes \
fillmiss -fillmiss \
fillmiss2 -fillmiss2 \
fldavg -fldavg \
fldcor -fldcor \
fldcount -fldcount \
fldcovar -fldcovar \
fldint -fldint \
fldkurt -fldkurt \
fldmax -fldmax \
fldmean -fldmean \
fldmedian -fldmedian \
fldmin -fldmin \
fldpctl -fldpctl \
fldrange -fldrange \
fldrms -fldrms \
fldskew -fldskew \
fldstd -fldstd \
fldstd1 -fldstd1 \
fldsum -fldsum \
fldvar -fldvar \
fldvar1 -fldvar1 \
for -for \
fourier -fourier \
fourier2grid -fourier2grid \
ge -ge \
gec -gec \
genbic -genbic \
genbil -genbil \
gencon -gencon \
gendis -gendis \
gengrid -gengrid \
genknn -genknn \
genlaf -genlaf \
genlevelbounds -genlevelbounds \
gennn -gennn \
genycon -genycon \
genycon2test -genycon2test \
gh2hl -gh2hl \
gh2hlx -gh2hlx \
gheight -gheight \
gheight_full -gheight_full \
gheight_half -gheight_half \
gheighthalf -gheighthalf \
globavg -globavg \
gmtcells -gmtcells \
gmtxyz -gmtxyz \
gp2fc -gp2fc \
gp2sp -gp2sp \
gp2spl -gp2spl \
gradsdes -gradsdes \
graph -graph \
grfill -grfill \
grid2fourier -grid2fourier \
gridarea -gridarea \
gridboxavg -gridboxavg \
gridboxkurt -gridboxkurt \
gridboxmax -gridboxmax \
gridboxmean -gridboxmean \
gridboxmedian -gridboxmedian \
gridboxmin -gridboxmin \
gridboxrange -gridboxrange \
gridboxskew -gridboxskew \
gridboxstd -gridboxstd \
gridboxstd1 -gridboxstd1 \
gridboxsum -gridboxsum \
gridboxvar -gridboxvar \
gridboxvar1 -gridboxvar1 \
gridcellidx -gridcellidx \
gridcellindex -gridcellindex \
griddes -griddes \
griddes2 -griddes2 \
griddx -griddx \
griddy -griddy \
gridmask -gridmask \
gridweights -gridweights \
gt -gt \
gtc -gtc \
harmonic -harmonic \
hi -hi \
highpass -highpass \
histcount -histcount \
histfreq -histfreq \
histmean -histmean \
histsum -histsum \
houravg -houravg \
hourcount -hourcount \
hourmax -hourmax \
hourmean -hourmean \
hourmin -hourmin \
hourpctl -hourpctl \
hourrange -hourrange \
hourstd -hourstd \
hourstd1 -hourstd1 \
hoursum -hoursum \
hourvar -hourvar \
hourvar1 -hourvar1 \
hpdegrade -hpdegrade \
hpupgrade -hpupgrade \
hurr -hurr \
ifnotthen -ifnotthen \
ifnotthenc -ifnotthenc \
ifthen -ifthen \
ifthenc -ifthenc \
ifthenelse -ifthenelse \
im -im \
import_binary -import_binary \
import_cmsaf -import_cmsaf \
import_e5ml -import_e5ml \
import_fv3grid -import_fv3grid \
import_grads -import_grads \
import_obs -import_obs \
imtocomplex -imtocomplex \
info -info \
infoc -infoc \
infon -infon \
infop -infop \
infov -infov \
input -input \
inputext -inputext \
inputsrv -inputsrv \
int -int \
intgridbil -intgridbil \
intgriddis -intgriddis \
intgridknn -intgridknn \
intgridnn -intgridnn \
intgridtraj -intgridtraj \
intlevel -intlevel \
intlevel3d -intlevel3d \
intlevelx -intlevelx \
intlevelx3d -intlevelx3d \
intntime -intntime \
inttime -inttime \
intyear -intyear \
invertlat -invertlat \
invertlatdata -invertlatdata \
invertlatdes -invertlatdes \
invertlev -invertlev \
invertlon -invertlon \
invertlondata -invertlondata \
invertlondes -invertlondes \
isosurface -isosurface \
le -le \
lec -lec \
lic -lic \
linfo -linfo \
ln -ln \
log -log \
log10 -log10 \
lowpass -lowpass \
lt -lt \
ltc -ltc \
map -map \
mask -mask \
maskcircle -maskcircle \
maskindexbox -maskindexbox \
masklonlatbox -masklonlatbox \
maskregion -maskregion \
mastrfu -mastrfu \
max -max \
maxc -maxc \
meandiff2test -meandiff2test \
meravg -meravg \
merge -merge \
mergegrid -mergegrid \
mergetime -mergetime \
merkurt -merkurt \
mermax -mermax \
mermean -mermean \
mermedian -mermedian \
mermin -mermin \
merpctl -merpctl \
merrange -merrange \
merskew -merskew \
merstd -merstd \
merstd1 -merstd1 \
mersum -mersum \
mervar -mervar \
mervar1 -mervar1 \
min -min \
minc -minc \
ml2hl -ml2hl \
ml2hlx -ml2hlx \
ml2pl -ml2pl \
ml2plx -ml2plx \
mod -mod \
monadd -monadd \
monavg -monavg \
moncount -moncount \
mondiv -mondiv \
monmax -monmax \
monmean -monmean \
monmin -monmin \
monmul -monmul \
monpctl -monpctl \
monrange -monrange \
monstd -monstd \
monstd1 -monstd1 \
monsub -monsub \
monsum -monsum \
monvar -monvar \
monvar1 -monvar1 \
mrotuv -mrotuv \
mrotuvb -mrotuvb \
mul -mul \
mulc -mulc \
mulcoslat -mulcoslat \
muldoy -muldoy \
muldpm -muldpm \
muldpy -muldpy \
ncode -ncode \
ncopy -ncopy \
ndate -ndate \
ne -ne \
nec -nec \
ngridpoints -ngridpoints \
ngrids -ngrids \
nint -nint \
nlevel -nlevel \
nmon -nmon \
not -not \
npar -npar \
ntime -ntime \
nvar -nvar \
nyear -nyear \
output -output \
outputarr -outputarr \
outputbounds -outputbounds \
outputboundscpt -outputboundscpt \
outputcenter -outputcenter \
outputcenter2 -outputcenter2 \
outputcentercpt -outputcentercpt \
outputext -outputext \
outputf -outputf \
outputfld -outputfld \
outputint -outputint \
outputkey -outputkey \
outputkml -outputkml \
outputsrv -outputsrv \
outputtab -outputtab \
outputtri -outputtri \
outputts -outputts \
outputvector -outputvector \
outputvrml -outputvrml \
outputxyz -outputxyz \
pack -pack \
pardup -pardup \
parmul -parmul \
partab -partab \
partab2 -partab2 \
pinfo -pinfo \
pinfov -pinfov \
pow -pow \
pressure -pressure \
pressure_full -pressure_full \
pressure_half -pressure_half \
projuvLatLon -projuvLatLon \
query -query \
rand -rand \
random -random \
re -re \
reci -reci \
recttocomplex -recttocomplex \
reducegrid -reducegrid \
regres -regres \
remap -remap \
remapavg -remapavg \
remapavgtest -remapavgtest \
remapbic -remapbic \
remapbil -remapbil \
remapcon -remapcon \
remapdis -remapdis \
remapeta -remapeta \
remapeta_s -remapeta_s \
remapeta_z -remapeta_z \
remapknn -remapknn \
remapkurt -remapkurt \
remaplaf -remaplaf \
remapmax -remapmax \
remapmean -remapmean \
remapmedian -remapmedian \
remapmin -remapmin \
remapnn -remapnn \
remaprange -remaprange \
remapskew -remapskew \
remapstd -remapstd \
remapstd1 -remapstd1 \
remapsum -remapsum \
remapvar -remapvar \
remapvar1 -remapvar1 \
remapycon -remapycon \
remapycon2test -remapycon2test \
replace -replace \
retocomplex -retocomplex \
rhopot -rhopot \
rotuvN -rotuvN \
rotuvNorth -rotuvNorth \
rotuvb -rotuvb \
runavg -runavg \
runmax -runmax \
runmean -runmean \
runmin -runmin \
runpctl -runpctl \
runrange -runrange \
runstd -runstd \
runstd1 -runstd1 \
runsum -runsum \
runvar -runvar \
runvar1 -runvar1 \
samplegrid -samplegrid \
samplegridicon -samplegridicon \
sealevelpressure -sealevelpressure \
seasavg -seasavg \
seascount -seascount \
seasmax -seasmax \
seasmean -seasmean \
seasmin -seasmin \
seasmonavg -seasmonavg \
seasmonmean -seasmonmean \
seaspctl -seaspctl \
seasrange -seasrange \
seasstd -seasstd \
seasstd1 -seasstd1 \
seassum -seassum \
seasvar -seasvar \
seasvar1 -seasvar1 \
seinfo -seinfo \
seinfoc -seinfoc \
seinfon -seinfon \
seinfop -seinfop \
selcircle -selcircle \
selcode -selcode \
seldate -seldate \
selday -selday \
select -select \
selgrid -selgrid \
selgridcell -selgridcell \
selgridname -selgridname \
selhour -selhour \
selindexbox -selindexbox \
sellevel -sellevel \
sellevidx -sellevidx \
sellonlatbox -sellonlatbox \
selltype -selltype \
selmon -selmon \
selmonth -selmonth \
selmulti -selmulti \
selname -selname \
seloperator -seloperator \
selparam -selparam \
selrec -selrec \
selregion -selregion \
selseas -selseas \
selseason -selseason \
selsmon -selsmon \
selstdname -selstdname \
seltabnum -seltabnum \
seltime -seltime \
seltimeidx -seltimeidx \
seltimestep -seltimestep \
selvar -selvar \
selyear -selyear \
selyearidx -selyearidx \
selzaxis -selzaxis \
selzaxisname -selzaxisname \
seq -seq \
setattribute -setattribute \
setcalendar -setcalendar \
setchunkspec -setchunkspec \
setcindexbox -setcindexbox \
setclonlatbox -setclonlatbox \
setcode -setcode \
setcodetab -setcodetab \
setctomiss -setctomiss \
setdate -setdate \
setday -setday \
setfilter -setfilter \
setgrid -setgrid \
setgridarea -setgridarea \
setgridcell -setgridcell \
setgridmask -setgridmask \
setgridnumber -setgridnumber \
setgridtype -setgridtype \
setgriduri -setgriduri \
sethalo -sethalo \
setlevel -setlevel \
setltype -setltype \
setmaxsteps -setmaxsteps \
setmiss -setmiss \
setmisstoc -setmisstoc \
setmisstodis -setmisstodis \
setmisstonn -setmisstonn \
setmissval -setmissval \
setmon -setmon \
setname -setname \
setparam -setparam \
setpartab -setpartab \
setpartabc -setpartabc \
setpartabn -setpartabn \
setpartabp -setpartabp \
setpartabv -setpartabv \
setprojparams -setprojparams \
setrcaname -setrcaname \
setreftime -setreftime \
setrtoc -setrtoc \
setrtoc2 -setrtoc2 \
setrtomiss -setrtomiss \
setstdname -setstdname \
settabnum -settabnum \
settaxis -settaxis \
settbounds -settbounds \
settime -settime \
settunits -settunits \
setunit -setunit \
setvals -setvals \
setvar -setvar \
setvrange -setvrange \
setyear -setyear \
setzaxis -setzaxis \
shaded -shaded \
shifttime -shifttime \
shiftx -shiftx \
shifty -shifty \
showattribute -showattribute \
showattsvar -showattsvar \
showchunkspec -showchunkspec \
showcode -showcode \
showdate -showdate \
showfilter -showfilter \
showformat -showformat \
showgrid -showgrid \
showhistory -showhistory \
showlevel -showlevel \
showltype -showltype \
showmon -showmon \
showname -showname \
showparam -showparam \
showstdname -showstdname \
showtime -showtime \
showtimestamp -showtimestamp \
showunit -showunit \
showvar -showvar \
showyear -showyear \
sin -sin \
sincos -sincos \
sinfo -sinfo \
sinfoc -sinfoc \
sinfon -sinfon \
sinfop -sinfop \
sinfov -sinfov \
smooth -smooth \
smooth9 -smooth9 \
sortcode -sortcode \
sortlevel -sortlevel \
sortname -sortname \
sortparam -sortparam \
sorttaxis -sorttaxis \
sorttimestamp -sorttimestamp \
sortvar -sortvar \
sp2fc -sp2fc \
sp2gp -sp2gp \
sp2gpl -sp2gpl \
sp2sp -sp2sp \
spartab -spartab \
spcut -spcut \
specinfo -specinfo \
spectrum -spectrum \
splitcode -splitcode \
splitdate -splitdate \
splitdatetime -splitdatetime \
splitday -splitday \
splitensemble -splitensemble \
splitgrid -splitgrid \
splithour -splithour \
splitlevel -splitlevel \
splitmon -splitmon \
splitname -splitname \
splitparam -splitparam \
splitrec -splitrec \
splitseas -splitseas \
splitsel -splitsel \
splittabnum -splittabnum \
splitvar -splitvar \
splityear -splityear \
splityearmon -splityearmon \
splitzaxis -splitzaxis \
sqr -sqr \
sqrt -sqrt \
stdatm -stdatm \
strbre -strbre \
stream -stream \
strgal -strgal \
strwin -strwin \
sub -sub \
subc -subc \
subgrid -subgrid \
subtrend -subtrend \
symmetrize -symmetrize \
szip -szip \
tan -tan \
tee -tee \
temp -temp \
testcellsearch -testcellsearch \
testfield -testfield \
testpointsearch -testpointsearch \
thinout -thinout \
timavg -timavg \
timcor -timcor \
timcount -timcount \
timcovar -timcovar \
timcumsum -timcumsum \
timederivative -timederivative \
timfillmiss -timfillmiss \
timmax -timmax \
timmaxidx -timmaxidx \
timmean -timmean \
timmin -timmin \
timminidx -timminidx \
timpctl -timpctl \
timrange -timrange \
timrmsd -timrmsd \
timselavg -timselavg \
timselmax -timselmax \
timselmean -timselmean \
timselmin -timselmin \
timselpctl -timselpctl \
timselrange -timselrange \
timselstd -timselstd \
timselstd1 -timselstd1 \
timselsum -timselsum \
timselvar -timselvar \
timselvar1 -timselvar1 \
timsort -timsort \
timstd -timstd \
timstd1 -timstd1 \
timsum -timsum \
timvar -timvar \
timvar1 -timvar1 \
tinfo -tinfo \
topo -topo \
topvalue -topvalue \
tpnhalo -tpnhalo \
transxy -transxy \
trend -trend \
tstepcount -tstepcount \
unpack -unpack \
unsetgridmask -unsetgridmask \
usegridnumber -usegridnumber \
uv2dv -uv2dv \
uv2dv_cfd -uv2dv_cfd \
uv2dvl -uv2dvl \
uv2vr_cfd -uv2vr_cfd \
uvDestag -uvDestag \
vardes -vardes \
varquot2test -varquot2test \
varrms -varrms \
varsavg -varsavg \
varskurt -varskurt \
varsmax -varsmax \
varsmean -varsmean \
varsmedian -varsmedian \
varsmin -varsmin \
varspctl -varspctl \
varsrange -varsrange \
varsskew -varsskew \
varsstd -varsstd \
varsstd1 -varsstd1 \
varssum -varssum \
varsvar -varsvar \
varsvar1 -varsvar1 \
vct -vct \
vct2 -vct2 \
vector -vector \
verifygrid -verifygrid \
verifyweights -verifyweights \
vertavg -vertavg \
vertcum -vertcum \
vertcumhl -vertcumhl \
vertfillmiss -vertfillmiss \
vertint -vertint \
vertmax -vertmax \
vertmean -vertmean \
vertmin -vertmin \
vertrange -vertrange \
vertstd -vertstd \
vertstd1 -vertstd1 \
vertsum -vertsum \
vertvar -vertvar \
vertvar1 -vertvar1 \
vertwind -vertwind \
vinfo -vinfo \
vlist -vlist \
wct -wct \
writegrid -writegrid \
writerandom -writerandom \
writeremapscrip -writeremapscrip \
xsinfo -xsinfo \
xsinfoc -xsinfoc \
xsinfon -xsinfon \
xsinfop -xsinfop \
ydayadd -ydayadd \
ydayavg -ydayavg \
ydaydiv -ydaydiv \
ydaymax -ydaymax \
ydaymean -ydaymean \
ydaymin -ydaymin \
ydaymul -ydaymul \
ydaypctl -ydaypctl \
ydayrange -ydayrange \
ydaystd -ydaystd \
ydaystd1 -ydaystd1 \
ydaysub -ydaysub \
ydaysum -ydaysum \
ydayvar -ydayvar \
ydayvar1 -ydayvar1 \
ydrunavg -ydrunavg \
ydrunmax -ydrunmax \
ydrunmean -ydrunmean \
ydrunmin -ydrunmin \
ydrunpctl -ydrunpctl \
ydrunstd -ydrunstd \
ydrunstd1 -ydrunstd1 \
ydrunsum -ydrunsum \
ydrunvar -ydrunvar \
ydrunvar1 -ydrunvar1 \
yearadd -yearadd \
yearavg -yearavg \
yearcount -yearcount \
yeardiv -yeardiv \
yearmax -yearmax \
yearmaxidx -yearmaxidx \
yearmean -yearmean \
yearmin -yearmin \
yearminidx -yearminidx \
yearmonavg -yearmonavg \
yearmonmean -yearmonmean \
yearmul -yearmul \
yearpctl -yearpctl \
yearrange -yearrange \
yearstd -yearstd \
yearstd1 -yearstd1 \
yearsub -yearsub \
yearsum -yearsum \
yearvar -yearvar \
yearvar1 -yearvar1 \
yhouradd -yhouradd \
yhouravg -yhouravg \
yhourdiv -yhourdiv \
yhourmax -yhourmax \
yhourmean -yhourmean \
yhourmin -yhourmin \
yhourmul -yhourmul \
yhourrange -yhourrange \
yhourstd -yhourstd \
yhourstd1 -yhourstd1 \
yhoursub -yhoursub \
yhoursum -yhoursum \
yhourvar -yhourvar \
yhourvar1 -yhourvar1 \
ymonadd -ymonadd \
ymonavg -ymonavg \
ymondiv -ymondiv \
ymoneq -ymoneq \
ymonge -ymonge \
ymongt -ymongt \
ymonle -ymonle \
ymonlt -ymonlt \
ymonmax -ymonmax \
ymonmean -ymonmean \
ymonmin -ymonmin \
ymonmul -ymonmul \
ymonne -ymonne \
ymonpctl -ymonpctl \
ymonrange -ymonrange \
ymonstd -ymonstd \
ymonstd1 -ymonstd1 \
ymonsub -ymonsub \
ymonsum -ymonsum \
ymonvar -ymonvar \
ymonvar1 -ymonvar1 \
yseasadd -yseasadd \
yseasavg -yseasavg \
yseasdiv -yseasdiv \
yseaseq -yseaseq \
yseasge -yseasge \
yseasgt -yseasgt \
yseasle -yseasle \
yseaslt -yseaslt \
yseasmax -yseasmax \
yseasmean -yseasmean \
yseasmin -yseasmin \
yseasmul -yseasmul \
yseasne -yseasne \
yseaspctl -yseaspctl \
yseasrange -yseasrange \
yseasstd -yseasstd \
yseasstd1 -yseasstd1 \
yseassub -yseassub \
yseassum -yseassum \
yseasvar -yseasvar \
yseasvar1 -yseasvar1 \
zaxisdes -zaxisdes \
zonavg -zonavg \
zonkurt -zonkurt \
zonmax -zonmax \
zonmean -zonmean \
zonmedian -zonmedian \
zonmin -zonmin \
zonpctl -zonpctl \
zonrange -zonrange \
zonskew -zonskew \
zonstd -zonstd \
zonstd1 -zonstd1 \
zonsum -zonsum \
zonvar -zonvar \
zonvar1 -zonvar1 \
zs2zl -zs2zl \
zs2zlx -zs2zlx \
zsdepth -zsdepth \
)" -f cdo
