// Automatically created with man2help.sh, don't edit!

#ifndef OPERATOR_HELP_H
#define OPERATOR_HELP_H

#include <vector>
#include <string_view>

typedef std::vector<std::string_view> CdoHelp;

extern const CdoHelp AdisitHelp;
extern const CdoHelp AfterburnerHelp;
extern const CdoHelp ArithHelp;
extern const CdoHelp ArithcHelp;
extern const CdoHelp ArithdaysHelp;
extern const CdoHelp ArithlatHelp;
extern const CdoHelp BitroundingHelp;
extern const CdoHelp ChangeHelp;
extern const CdoHelp CmorHelp;
extern const CdoHelp CmorliteHelp;
extern const CdoHelp CollgridHelp;
extern const CdoHelp CompHelp;
extern const CdoHelp CompcHelp;
extern const CdoHelp CondHelp;
extern const CdoHelp Cond2Help;
extern const CdoHelp CondcHelp;
extern const CdoHelp ConsecstatHelp;
extern const CdoHelp CopyHelp;
extern const CdoHelp DayarithHelp;
extern const CdoHelp DaypctlHelp;
extern const CdoHelp DaystatHelp;
extern const CdoHelp DeltatHelp;
extern const CdoHelp DeriveparHelp;
extern const CdoHelp DetrendHelp;
extern const CdoHelp DhourstatHelp;
extern const CdoHelp DiffHelp;
extern const CdoHelp DistgridHelp;
extern const CdoHelp DminutestatHelp;
extern const CdoHelp DuplicateHelp;
extern const CdoHelp Eca_cddHelp;
extern const CdoHelp Eca_cfdHelp;
extern const CdoHelp Eca_csuHelp;
extern const CdoHelp Eca_cwdHelp;
extern const CdoHelp Eca_cwdiHelp;
extern const CdoHelp Eca_cwfiHelp;
extern const CdoHelp Eca_etccdiHelp;
extern const CdoHelp Eca_etrHelp;
extern const CdoHelp Eca_fdHelp;
extern const CdoHelp Eca_gslHelp;
extern const CdoHelp Eca_hdHelp;
extern const CdoHelp Eca_hwdiHelp;
extern const CdoHelp Eca_hwfiHelp;
extern const CdoHelp Eca_idHelp;
extern const CdoHelp Eca_pdHelp;
extern const CdoHelp Eca_r75pHelp;
extern const CdoHelp Eca_r75ptotHelp;
extern const CdoHelp Eca_r90pHelp;
extern const CdoHelp Eca_r90ptotHelp;
extern const CdoHelp Eca_r95pHelp;
extern const CdoHelp Eca_r95ptotHelp;
extern const CdoHelp Eca_r99pHelp;
extern const CdoHelp Eca_r99ptotHelp;
extern const CdoHelp Eca_rr1Help;
extern const CdoHelp Eca_rx1dayHelp;
extern const CdoHelp Eca_rx5dayHelp;
extern const CdoHelp Eca_sdiiHelp;
extern const CdoHelp Eca_suHelp;
extern const CdoHelp Eca_tg10pHelp;
extern const CdoHelp Eca_tg90pHelp;
extern const CdoHelp Eca_tn10pHelp;
extern const CdoHelp Eca_tn90pHelp;
extern const CdoHelp Eca_trHelp;
extern const CdoHelp Eca_tx10pHelp;
extern const CdoHelp Eca_tx90pHelp;
extern const CdoHelp EnlargeHelp;
extern const CdoHelp EnsstatHelp;
extern const CdoHelp Ensstat2Help;
extern const CdoHelp EnsvalHelp;
extern const CdoHelp EofHelp;
extern const CdoHelp EofcoeffHelp;
extern const CdoHelp ExprHelp;
extern const CdoHelp FdnsHelp;
extern const CdoHelp FiledesHelp;
extern const CdoHelp FilterHelp;
extern const CdoHelp FldcorHelp;
extern const CdoHelp FldcovarHelp;
extern const CdoHelp FldstatHelp;
extern const CdoHelp FourierHelp;
extern const CdoHelp GetgridcellHelp;
extern const CdoHelp GradsdesHelp;
extern const CdoHelp GridboxstatHelp;
extern const CdoHelp GridcellHelp;
extern const CdoHelp HealpixHelp;
extern const CdoHelp HistogramHelp;
extern const CdoHelp HourpctlHelp;
extern const CdoHelp HourstatHelp;
extern const CdoHelp HurrHelp;
extern const CdoHelp ImportbinaryHelp;
extern const CdoHelp ImportcmsafHelp;
extern const CdoHelp InfoHelp;
extern const CdoHelp InputHelp;
extern const CdoHelp IntlevelHelp;
extern const CdoHelp Intlevel3dHelp;
extern const CdoHelp InttimeHelp;
extern const CdoHelp IntyearHelp;
extern const CdoHelp InvertHelp;
extern const CdoHelp InvertlevHelp;
extern const CdoHelp MaggraphHelp;
extern const CdoHelp MagplotHelp;
extern const CdoHelp MagvectorHelp;
extern const CdoHelp MapreduceHelp;
extern const CdoHelp MaskboxHelp;
extern const CdoHelp MaskregionHelp;
extern const CdoHelp MastrfuHelp;
extern const CdoHelp MathHelp;
extern const CdoHelp MergeHelp;
extern const CdoHelp MergegridHelp;
extern const CdoHelp MerstatHelp;
extern const CdoHelp MonarithHelp;
extern const CdoHelp MonpctlHelp;
extern const CdoHelp MonstatHelp;
extern const CdoHelp MrotuvbHelp;
extern const CdoHelp Ncl_windHelp;
extern const CdoHelp NinfoHelp;
extern const CdoHelp OutputHelp;
extern const CdoHelp OutputgmtHelp;
extern const CdoHelp OutputtabHelp;
extern const CdoHelp PackHelp;
extern const CdoHelp PressureHelp;
extern const CdoHelp RegresHelp;
extern const CdoHelp RemapHelp;
extern const CdoHelp RemapbicHelp;
extern const CdoHelp RemapbilHelp;
extern const CdoHelp RemapconHelp;
extern const CdoHelp RemapdisHelp;
extern const CdoHelp RemapetaHelp;
extern const CdoHelp RemapknnHelp;
extern const CdoHelp RemaplafHelp;
extern const CdoHelp RemapnnHelp;
extern const CdoHelp RemapstatHelp;
extern const CdoHelp ReplaceHelp;
extern const CdoHelp ReplacevaluesHelp;
extern const CdoHelp RhopotHelp;
extern const CdoHelp RotuvHelp;
extern const CdoHelp RunpctlHelp;
extern const CdoHelp RunstatHelp;
extern const CdoHelp SamplegridHelp;
extern const CdoHelp SeaspctlHelp;
extern const CdoHelp SeasstatHelp;
extern const CdoHelp SelboxHelp;
extern const CdoHelp SelectHelp;
extern const CdoHelp SelgridcellHelp;
extern const CdoHelp SelmultiHelp;
extern const CdoHelp SelregionHelp;
extern const CdoHelp SelsurfaceHelp;
extern const CdoHelp SeltimeHelp;
extern const CdoHelp SeltimeidxHelp;
extern const CdoHelp SelvarHelp;
extern const CdoHelp SelyearidxHelp;
extern const CdoHelp SetHelp;
extern const CdoHelp SetattributeHelp;
extern const CdoHelp SetboxHelp;
extern const CdoHelp SetchunkspecHelp;
extern const CdoHelp SetfilterHelp;
extern const CdoHelp SetgridHelp;
extern const CdoHelp SetgridcellHelp;
extern const CdoHelp SethaloHelp;
extern const CdoHelp SetmissHelp;
extern const CdoHelp SetpartabHelp;
extern const CdoHelp SettimeHelp;
extern const CdoHelp SetzaxisHelp;
extern const CdoHelp ShiftxyHelp;
extern const CdoHelp ShowattributeHelp;
extern const CdoHelp ShowinfoHelp;
extern const CdoHelp SinfoHelp;
extern const CdoHelp SmoothHelp;
extern const CdoHelp SpecconvHelp;
extern const CdoHelp SpectralHelp;
extern const CdoHelp SplitHelp;
extern const CdoHelp SplitdateHelp;
extern const CdoHelp SplitselHelp;
extern const CdoHelp SplittimeHelp;
extern const CdoHelp StrbreHelp;
extern const CdoHelp StrgalHelp;
extern const CdoHelp StrwinHelp;
extern const CdoHelp SymmetrizeHelp;
extern const CdoHelp TeeHelp;
extern const CdoHelp TimcorHelp;
extern const CdoHelp TimcovarHelp;
extern const CdoHelp TimcumsumHelp;
extern const CdoHelp TimfillmissHelp;
extern const CdoHelp TimpctlHelp;
extern const CdoHelp TimselpctlHelp;
extern const CdoHelp TimselstatHelp;
extern const CdoHelp TimsortHelp;
extern const CdoHelp TimstatHelp;
extern const CdoHelp TrendHelp;
extern const CdoHelp TrendarithHelp;
extern const CdoHelp UnpackHelp;
extern const CdoHelp VargenHelp;
extern const CdoHelp VarsstatHelp;
extern const CdoHelp VerifygridHelp;
extern const CdoHelp VertfillmissHelp;
extern const CdoHelp VertintapHelp;
extern const CdoHelp VertintghHelp;
extern const CdoHelp VertintmlHelp;
extern const CdoHelp VertstatHelp;
extern const CdoHelp WctHelp;
extern const CdoHelp WindHelp;
extern const CdoHelp Wind2Help;
extern const CdoHelp WindtransHelp;
extern const CdoHelp XsinfoHelp;
extern const CdoHelp YdayarithHelp;
extern const CdoHelp YdaypctlHelp;
extern const CdoHelp YdaystatHelp;
extern const CdoHelp YdrunpctlHelp;
extern const CdoHelp YdrunstatHelp;
extern const CdoHelp YeararithHelp;
extern const CdoHelp YearmonstatHelp;
extern const CdoHelp YearpctlHelp;
extern const CdoHelp YearstatHelp;
extern const CdoHelp YhourarithHelp;
extern const CdoHelp YhourstatHelp;
extern const CdoHelp YmonarithHelp;
extern const CdoHelp YmoncompHelp;
extern const CdoHelp YmonpctlHelp;
extern const CdoHelp YmonstatHelp;
extern const CdoHelp YseasarithHelp;
extern const CdoHelp YseascompHelp;
extern const CdoHelp YseaspctlHelp;
extern const CdoHelp YseasstatHelp;
extern const CdoHelp ZonstatHelp;

#endif
