#include <stdio.h>
#include <stdlib.h>
#include "jsmn.h"

static
void jsmn_init(jsmn_parser *parser)
{
  parser->tokens = NULL;
  parser->num_tokens = 0;
  parser->toksuper = -1;
  parser->toknext = 0;
  parser->pos = 0;
  parser->lineno = 0;
}


jsmn_parser *jsmn_new(void)
{
  jsmn_parser *parser = (jsmn_parser *) malloc(sizeof(jsmn_parser));
  jsmn_init(parser);

  return parser;
}


void jsmn_destroy(jsmn_parser *parser)
{
  if ( parser )
    {
      if ( parser->tokens ) free(parser->tokens);
      jsmn_init(parser);
      free(parser);
    }
}

/**
 * Allocates a fresh unused token from the token pull.
 */
static
jsmntok_t *jsmn_alloc_token(jsmn_parser *parser)
{
  const unsigned int TOK_MEM_INCR = 64;

  if ( parser->toknext >= parser->num_tokens )
    {
      parser->num_tokens += TOK_MEM_INCR;
      parser->tokens = (jsmntok_t *) realloc(parser->tokens, sizeof(jsmntok_t) * parser->num_tokens);
      if ( parser->tokens == NULL )
        {
          fprintf(stderr, "%s: Failed to allocated more memory!", __func__);
          exit(-1);
        }
    }

  jsmntok_t *tok = &parser->tokens[parser->toknext++];
  tok->start = tok->end = -1;
  tok->size = 0;
#ifdef JSMN_PARENT_LINKS
  tok->parent = -1;
#endif
  
  return tok;
}

/**
 * Fills token type and boundaries.
 */
static
void jsmn_fill_token(jsmntok_t *token, jsmntype_t type, int start, int end)
{
  token->type = type;
  token->start = start;
  token->end = end;
  token->size = 0;
}

// Fills next available token with JSON primitive.
static
int jsmn_parse_primitive(jsmn_parser *parser, const char *js, size_t len)
{
  jsmntok_t *token;
  int start = parser->pos;

  for (; parser->pos < len && js[parser->pos] != '\0'; parser->pos++)
    {
      switch (js[parser->pos])
        {
#ifndef JSMN_STRICT
        /* In strict mode primitive must be followed by "," or "}" or "]" */
        case ':':
#endif
        case '\t' : case '\r' : case '\n' : case ' ' :
        case ','  : case ']'  : case '}' :
          goto found;
        }
      if (js[parser->pos] < 32 || js[parser->pos] >= 127)
        {
          parser->pos = start;
          return JSMN_ERROR_INVAL;
        }
    }
#ifdef JSMN_STRICT
  /* In strict mode primitive must be followed by a comma/object/array */
  parser->pos = start;
  return JSMN_ERROR_PART;
#endif

found:

  token = jsmn_alloc_token(parser);
  jsmn_fill_token(token, JSMN_PRIMITIVE, start, parser->pos);
#ifdef JSMN_PARENT_LINKS
  token->parent = parser->toksuper;
#endif
  parser->pos--;

  return 0;
}

/**
 * Fills next token with JSON string.
 */
static int jsmn_parse_string(jsmn_parser *parser, const char *js, size_t len)
{
  jsmntok_t *token;
  int start = parser->pos;

  parser->pos++;

  /* Skip starting quote */
  for (; parser->pos < len && js[parser->pos] != '\0'; parser->pos++)
    {
      char c = js[parser->pos];

      /* Quote: end of string */
      if (c == '\"')
        {
          token = jsmn_alloc_token(parser);
          jsmn_fill_token(token, JSMN_STRING, start+1, parser->pos);
#ifdef JSMN_PARENT_LINKS
          token->parent = parser->toksuper;
#endif
          return 0;
        }

      /* Backslash: Quoted symbol expected */
      if (c == '\\' && parser->pos + 1 < len)
        {
          parser->pos++;
          switch (js[parser->pos]) {
            /* Allowed escaped symbols */
          case '\"': case '/' : case '\\' : case 'b' :
          case 'f' : case 'r' : case 'n'  : case 't' :
            break;
            /* Allows escaped symbol \uXXXX */
          case 'u':
            parser->pos++;
            for(int i = 0; i < 4 && parser->pos < len && js[parser->pos] != '\0'; i++)
              {
                /* If it isn't a hex character we have an error */
                if(!((js[parser->pos] >= 48 && js[parser->pos] <= 57) || /* 0-9 */
                     (js[parser->pos] >= 65 && js[parser->pos] <= 70) || /* A-F */
                     (js[parser->pos] >= 97 && js[parser->pos] <= 102))) { /* a-f */
                  parser->pos = start;
                  return JSMN_ERROR_INVAL;
                }
                parser->pos++;
              }
            parser->pos--;
            break;
            /* Unexpected symbol */
          default:
            parser->pos = start;
            return JSMN_ERROR_INVAL;
          }
        }
    }
  parser->pos = start;
  return JSMN_ERROR_PART;
}

// Parse JSON string and fill tokens.
int jsmn_parse(jsmn_parser *parser, const char *js, size_t len)
{
  int r;
  int i;
  jsmntok_t *token;

  parser->lineno = 1;

  for (; parser->pos < len && js[parser->pos] != '\0'; parser->pos++)
    {
      jsmntype_t type;

      char c = js[parser->pos];
      switch (c)
        {
        case '{': case '[':
          token = jsmn_alloc_token(parser);
          if (parser->toksuper != -1)
            {
              parser->tokens[parser->toksuper].size++;
#ifdef JSMN_PARENT_LINKS
              token->parent = parser->toksuper;
#endif
            }
          token->type = (c == '{' ? JSMN_OBJECT : JSMN_ARRAY);
          token->start = parser->pos;
          parser->toksuper = parser->toknext - 1;
          break;
        case '}': case ']':
          type = (c == '}' ? JSMN_OBJECT : JSMN_ARRAY);
#ifdef JSMN_PARENT_LINKS
          if (parser->toknext < 1) return JSMN_ERROR_INVAL;
          token = &parser->tokens[parser->toknext - 1];
          for (;;)
            {
              if (token->start != -1 && token->end == -1)
                {
                  if (token->type != type) return JSMN_ERROR_INVAL;
                  token->end = parser->pos + 1;
                  parser->toksuper = token->parent;
                  break;
                }
              if (token->parent == -1)
                {
                  if (token->type != type || parser->toksuper == -1) return JSMN_ERROR_INVAL;
                  break;
                }
              token = &parser->tokens[token->parent];
            }
#else
          for (i = parser->toknext - 1; i >= 0; i--)
            {
              token = &parser->tokens[i];
              if (token->start != -1 && token->end == -1)
                {
                  if (token->type != type) return JSMN_ERROR_INVAL;
                  parser->toksuper = -1;
                  token->end = parser->pos + 1;
                  break;
                }
            }
          /* Error if unmatched closing bracket */
          if (i == -1) return JSMN_ERROR_INVAL;
          for (; i >= 0; i--)
            {
              token = &parser->tokens[i];
              if (token->start != -1 && token->end == -1) {
                parser->toksuper = i;
                break;
              }
            }
#endif
          break;
        case '\"':
          r = jsmn_parse_string(parser, js, len);
          if (r < 0) return r;
          if ( parser->toksuper != -1 )
            parser->tokens[parser->toksuper].size++;
          break;
        case '\r':
          if ( parser->pos+1 < len && js[parser->pos+1] == '\n' ) parser->pos++;
          // C++20 [[fallthrough]];
        case '\n':
          parser->lineno++;
          break;
        case '\t': case ' ':
          break;
        case ':':
          parser->toksuper = parser->toknext - 1;
          break;
        case ',':
          if (parser->toksuper != -1 &&
              parser->tokens[parser->toksuper].type != JSMN_ARRAY &&
              parser->tokens[parser->toksuper].type != JSMN_OBJECT)
            {
#ifdef JSMN_PARENT_LINKS
              parser->toksuper = parser->tokens[parser->toksuper].parent;
#else
              for (i = parser->toknext - 1; i >= 0; i--)
                {
                  if (parser->tokens[i].type == JSMN_ARRAY || parser->tokens[i].type == JSMN_OBJECT)
                    {
                      if (parser->tokens[i].start != -1 && parser->tokens[i].end == -1)
                        {
                          parser->toksuper = i;
                          break;
                        }
                    }
                }
#endif
            }
          break;
#ifdef JSMN_STRICT
          /* In strict mode primitives are: numbers and booleans */
        case '-': case '0': case '1' : case '2': case '3' : case '4':
        case '5': case '6': case '7' : case '8': case '9':
        case 't': case 'f': case 'n' :
          /* And they must not be keys of the object */
          if (parser->toksuper != -1)
            {
              jsmntok_t *t = &parser->tokens[parser->toksuper];
              if (t->type == JSMN_OBJECT || (t->type == JSMN_STRING && t->size != 0))
                return JSMN_ERROR_INVAL;
            }
#else
          /* In non-strict mode every unquoted value is a primitive */
        default:
#endif
          r = jsmn_parse_primitive(parser, js, len);
          if (r < 0) return r;
          if ( parser->toksuper != -1 )
            parser->tokens[parser->toksuper].size++;
          break;

#ifdef JSMN_STRICT
          /* Unexpected char in strict mode */
        default:
          return JSMN_ERROR_INVAL;
#endif
        }
    }

  for (i = parser->toknext - 1; i >= 0; i--)
    {
      /* Unmatched opened object or array */
      if (parser->tokens[i].start != -1 && parser->tokens[i].end == -1) return JSMN_ERROR_PART;
    }

  return 0;
}
