#ifndef PIO_CLIENT_H
#define PIO_CLIENT_H

#include <pio_conf.h>

void cdiPioClientSetup(int pioNamespace, struct cdiPioConf *conf);

#endif

/*
 * Local Variables:
 * c-file-style: "Java"
 * c-basic-offset: 2
 * indent-tabs-mode: nil
 * show-trailing-whitespace: t
 * require-trailing-newline: t
 * End:
 */
