set cdoCmpl = (\
-apply \
-argument_groups \
-async_read \
-attribs \
-chunksize \
-chunkspec \
-cmor \
-config \
-copy_chunkspec \
-disable_file_suffix \
-disable_history \
-double \
-download_path \
-eccodes \
-enableexcept \
-envvars \
-file_suffix \
-force \
-gridsearchradius \
-history \
-history_info \
-icon_grids \
-ignore_time_bounds \
-module_info \
-netcdf_hdr_pad \
-no_history \
-no_remap_weights \
-nofile \
-operators \
-operators_no_output \
-pedantic \
-percentile \
-precision \
-print_filename \
-query \
-reduce_dim \
-remap_weights \
-remove_chunkspec \
-reset_history \
-rusage \
-seed \
-settings \
-shuffle \
-single \
-sortparam \
-test \
-timestat_date \
-use_fftw \
-use_time_bounds \
-version_info \
-worker \
A, \
C, \
D, \
F, \
M, \
O, \
P, \
Q, \
R, \
S, \
T, \
V, \
Z, \
a, \
b, \
c, \
d, \
f, \
g, \
h, \
i, \
k, \
l, \
m, \
p, \
r, \
s, \
t, \
u, \
v, \
w, \
z, \
abs \
acos \
add \
addc \
addtrend \
adipot \
adisit \
aexpr \
aexprf \
after \
afterburner \
air_density \
anomaly \
ap2hl \
ap2hlx \
ap2pl \
ap2plx \
arg \
asin \
atan \
atan2 \
bandpass \
bitrounding \
bottomvalue \
boxavg \
cat \
cdiread \
cdiwrite \
changemulti \
chcode \
chlevel \
chlevelc \
chlevelv \
chltype \
chname \
chparam \
chtabnum \
chunit \
chvar \
cinfo \
clone \
cloudlayer \
cmor \
cmorlite \
codetab \
collgrid \
complextopol \
complextorect \
conj \
consecsum \
consects \
const \
contour \
conv_cmor_table \
copy \
cos \
coshill \
dayadd \
dayavg \
daycount \
daydiv \
daymax \
daymean \
daymin \
daymul \
daypctl \
dayrange \
daystd \
daystd1 \
daysub \
daysum \
dayvar \
dayvar1 \
dcw \
del29feb \
delattribute \
delcode \
delday \
delete \
delgridcell \
delmulti \
delname \
delparam \
delta_pressure \
deltat \
delvar \
detrend \
dhouravg \
dhourmax \
dhourmean \
dhourmin \
dhourrange \
dhourstd \
dhourstd1 \
dhoursum \
dhourvar \
dhourvar1 \
diff \
diffc \
diffn \
diffp \
difftest \
diffv \
distgrid \
div \
divc \
divcoslat \
divdpm \
divdpy \
dminuteavg \
dminutemax \
dminutemean \
dminutemin \
dminuterange \
dminutestd \
dminutestd1 \
dminutesum \
dminutevar \
dminutevar1 \
dump_cmor_table \
dumpmap \
duplicate \
dv2ps \
dv2uv \
dv2uvl \
eca_cdd \
eca_cfd \
eca_csu \
eca_cwd \
eca_cwdi \
eca_cwfi \
eca_etr \
eca_fd \
eca_gsl \
eca_hd \
eca_hwdi \
eca_hwfi \
eca_id \
eca_pd \
eca_r10mm \
eca_r1mm \
eca_r20mm \
eca_r75p \
eca_r75ptot \
eca_r90p \
eca_r90ptot \
eca_r95p \
eca_r95ptot \
eca_r99p \
eca_r99ptot \
eca_rr1 \
eca_rx1day \
eca_rx5day \
eca_sdii \
eca_su \
eca_tg10p \
eca_tg90p \
eca_tn10p \
eca_tn90p \
eca_tr \
eca_tx10p \
eca_tx90p \
enlarge \
ensavg \
ensbrs \
enscrps \
enskurt \
ensmax \
ensmean \
ensmedian \
ensmin \
enspctl \
ensrange \
ensrkhistspace \
ensrkhisttime \
ensroc \
ensskew \
ensstd \
ensstd1 \
enssum \
ensvar \
ensvar1 \
eof \
eof3d \
eof3dspatial \
eof3dtime \
eofcoeff \
eofcoeff3d \
eofspatial \
eoftime \
eq \
eqc \
etccdi \
etccdi_cdd \
etccdi_csdi \
etccdi_cwd \
etccdi_fd \
etccdi_gsl \
etccdi_hd \
etccdi_id \
etccdi_r10mm \
etccdi_r1mm \
etccdi_r20mm \
etccdi_r95p \
etccdi_r99p \
etccdi_rx1day \
etccdi_rx1daymon \
etccdi_rx5day \
etccdi_rx5daymon \
etccdi_sdii \
etccdi_su \
etccdi_tn10p \
etccdi_tn90p \
etccdi_tr \
etccdi_tx10p \
etccdi_tx90p \
etccdi_wsdi \
exp \
export_e5ml \
expr \
exprf \
fc2gp \
fc2sp \
fdns \
filedes \
fillmiss \
fillmiss2 \
fldavg \
fldcor \
fldcount \
fldcovar \
fldint \
fldkurt \
fldmax \
fldmean \
fldmedian \
fldmin \
fldpctl \
fldrange \
fldrms \
fldskew \
fldstd \
fldstd1 \
fldsum \
fldvar \
fldvar1 \
for \
fourier \
fourier2grid \
ge \
gec \
genbic \
genbil \
gencon \
gendis \
gengrid \
genknn \
genlaf \
genlevelbounds \
gennn \
genycon \
genycon2test \
gh2hl \
gh2hlx \
gheight \
gheight_full \
gheight_half \
gheighthalf \
globavg \
gmtcells \
gmtxyz \
gp2fc \
gp2sp \
gp2spl \
gradsdes \
graph \
grfill \
grid2fourier \
gridarea \
gridboxavg \
gridboxkurt \
gridboxmax \
gridboxmean \
gridboxmedian \
gridboxmin \
gridboxrange \
gridboxskew \
gridboxstd \
gridboxstd1 \
gridboxsum \
gridboxvar \
gridboxvar1 \
gridcellidx \
gridcellindex \
griddes \
griddes2 \
griddx \
griddy \
gridmask \
gridweights \
gt \
gtc \
harmonic \
hi \
highpass \
histcount \
histfreq \
histmean \
histsum \
houravg \
hourcount \
hourmax \
hourmean \
hourmin \
hourpctl \
hourrange \
hourstd \
hourstd1 \
hoursum \
hourvar \
hourvar1 \
hpdegrade \
hpupgrade \
hurr \
ifnotthen \
ifnotthenc \
ifthen \
ifthenc \
ifthenelse \
im \
import_binary \
import_cmsaf \
import_e5ml \
import_fv3grid \
import_grads \
import_obs \
imtocomplex \
info \
infoc \
infon \
infop \
infov \
input \
inputext \
inputsrv \
int \
intgridbil \
intgriddis \
intgridknn \
intgridnn \
intgridtraj \
intlevel \
intlevel3d \
intlevelx \
intlevelx3d \
intntime \
inttime \
intyear \
invertlat \
invertlatdata \
invertlatdes \
invertlev \
invertlon \
invertlondata \
invertlondes \
isosurface \
le \
lec \
lic \
linfo \
ln \
log \
log10 \
lowpass \
lt \
ltc \
map \
mask \
maskcircle \
maskindexbox \
masklonlatbox \
maskregion \
mastrfu \
max \
maxc \
meandiff2test \
meravg \
merge \
mergegrid \
mergetime \
merkurt \
mermax \
mermean \
mermedian \
mermin \
merpctl \
merrange \
merskew \
merstd \
merstd1 \
mersum \
mervar \
mervar1 \
min \
minc \
ml2hl \
ml2hlx \
ml2pl \
ml2plx \
mod \
monadd \
monavg \
moncount \
mondiv \
monmax \
monmean \
monmin \
monmul \
monpctl \
monrange \
monstd \
monstd1 \
monsub \
monsum \
monvar \
monvar1 \
mrotuv \
mrotuvb \
mul \
mulc \
mulcoslat \
muldoy \
muldpm \
muldpy \
ncode \
ncopy \
ndate \
ne \
nec \
ngridpoints \
ngrids \
nint \
nlevel \
nmon \
not \
npar \
ntime \
nvar \
nyear \
output \
outputarr \
outputbounds \
outputboundscpt \
outputcenter \
outputcenter2 \
outputcentercpt \
outputext \
outputf \
outputfld \
outputint \
outputkey \
outputkml \
outputsrv \
outputtab \
outputtri \
outputts \
outputvector \
outputvrml \
outputxyz \
pack \
pardup \
parmul \
partab \
partab2 \
pinfo \
pinfov \
pow \
pressure \
pressure_full \
pressure_half \
projuvLatLon \
query \
rand \
random \
re \
reci \
recttocomplex \
reducegrid \
regres \
remap \
remapavg \
remapavgtest \
remapbic \
remapbil \
remapcon \
remapdis \
remapeta \
remapeta_s \
remapeta_z \
remapknn \
remapkurt \
remaplaf \
remapmax \
remapmean \
remapmedian \
remapmin \
remapnn \
remaprange \
remapskew \
remapstd \
remapstd1 \
remapsum \
remapvar \
remapvar1 \
remapycon \
remapycon2test \
replace \
retocomplex \
rhopot \
rotuvN \
rotuvNorth \
rotuvb \
runavg \
runmax \
runmean \
runmin \
runpctl \
runrange \
runstd \
runstd1 \
runsum \
runvar \
runvar1 \
samplegrid \
samplegridicon \
sealevelpressure \
seasavg \
seascount \
seasmax \
seasmean \
seasmin \
seasmonavg \
seasmonmean \
seaspctl \
seasrange \
seasstd \
seasstd1 \
seassum \
seasvar \
seasvar1 \
seinfo \
seinfoc \
seinfon \
seinfop \
selcircle \
selcode \
seldate \
selday \
select \
selgrid \
selgridcell \
selgridname \
selhour \
selindexbox \
sellevel \
sellevidx \
sellonlatbox \
selltype \
selmon \
selmonth \
selmulti \
selname \
seloperator \
selparam \
selrec \
selregion \
selseas \
selseason \
selsmon \
selstdname \
seltabnum \
seltime \
seltimeidx \
seltimestep \
selvar \
selyear \
selyearidx \
selzaxis \
selzaxisname \
seq \
setattribute \
setcalendar \
setchunkspec \
setcindexbox \
setclonlatbox \
setcode \
setcodetab \
setctomiss \
setdate \
setday \
setfilter \
setgrid \
setgridarea \
setgridcell \
setgridmask \
setgridnumber \
setgridtype \
setgriduri \
sethalo \
setlevel \
setltype \
setmaxsteps \
setmiss \
setmisstoc \
setmisstodis \
setmisstonn \
setmissval \
setmon \
setname \
setparam \
setpartab \
setpartabc \
setpartabn \
setpartabp \
setpartabv \
setprojparams \
setrcaname \
setreftime \
setrtoc \
setrtoc2 \
setrtomiss \
setstdname \
settabnum \
settaxis \
settbounds \
settime \
settunits \
setunit \
setvals \
setvar \
setvrange \
setyear \
setzaxis \
shaded \
shifttime \
shiftx \
shifty \
showattribute \
showattsvar \
showchunkspec \
showcode \
showdate \
showfilter \
showformat \
showgrid \
showhistory \
showlevel \
showltype \
showmon \
showname \
showparam \
showstdname \
showtime \
showtimestamp \
showunit \
showvar \
showyear \
sin \
sincos \
sinfo \
sinfoc \
sinfon \
sinfop \
sinfov \
smooth \
smooth9 \
sortcode \
sortlevel \
sortname \
sortparam \
sorttaxis \
sorttimestamp \
sortvar \
sp2fc \
sp2gp \
sp2gpl \
sp2sp \
spartab \
spcut \
specinfo \
spectrum \
splitcode \
splitdate \
splitdatetime \
splitday \
splitensemble \
splitgrid \
splithour \
splitlevel \
splitmon \
splitname \
splitparam \
splitrec \
splitseas \
splitsel \
splittabnum \
splitvar \
splityear \
splityearmon \
splitzaxis \
sqr \
sqrt \
stdatm \
strbre \
stream \
strgal \
strwin \
sub \
subc \
subgrid \
subtrend \
symmetrize \
szip \
tan \
tee \
temp \
testcellsearch \
testfield \
testpointsearch \
thinout \
timavg \
timcor \
timcount \
timcovar \
timcumsum \
timederivative \
timfillmiss \
timmax \
timmaxidx \
timmean \
timmin \
timminidx \
timpctl \
timrange \
timrmsd \
timselavg \
timselmax \
timselmean \
timselmin \
timselpctl \
timselrange \
timselstd \
timselstd1 \
timselsum \
timselvar \
timselvar1 \
timsort \
timstd \
timstd1 \
timsum \
timvar \
timvar1 \
tinfo \
topo \
topvalue \
tpnhalo \
transxy \
trend \
tstepcount \
unpack \
unsetgridmask \
usegridnumber \
uv2dv \
uv2dv_cfd \
uv2dvl \
uv2vr_cfd \
uvDestag \
vardes \
varquot2test \
varrms \
varsavg \
varskurt \
varsmax \
varsmean \
varsmedian \
varsmin \
varspctl \
varsrange \
varsskew \
varsstd \
varsstd1 \
varssum \
varsvar \
varsvar1 \
vct \
vct2 \
vector \
verifygrid \
verifyweights \
vertavg \
vertcum \
vertcumhl \
vertfillmiss \
vertint \
vertmax \
vertmean \
vertmin \
vertrange \
vertstd \
vertstd1 \
vertsum \
vertvar \
vertvar1 \
vertwind \
vinfo \
vlist \
wct \
writegrid \
writerandom \
writeremapscrip \
xsinfo \
xsinfoc \
xsinfon \
xsinfop \
ydayadd \
ydayavg \
ydaydiv \
ydaymax \
ydaymean \
ydaymin \
ydaymul \
ydaypctl \
ydayrange \
ydaystd \
ydaystd1 \
ydaysub \
ydaysum \
ydayvar \
ydayvar1 \
ydrunavg \
ydrunmax \
ydrunmean \
ydrunmin \
ydrunpctl \
ydrunstd \
ydrunstd1 \
ydrunsum \
ydrunvar \
ydrunvar1 \
yearadd \
yearavg \
yearcount \
yeardiv \
yearmax \
yearmaxidx \
yearmean \
yearmin \
yearminidx \
yearmonavg \
yearmonmean \
yearmul \
yearpctl \
yearrange \
yearstd \
yearstd1 \
yearsub \
yearsum \
yearvar \
yearvar1 \
yhouradd \
yhouravg \
yhourdiv \
yhourmax \
yhourmean \
yhourmin \
yhourmul \
yhourrange \
yhourstd \
yhourstd1 \
yhoursub \
yhoursum \
yhourvar \
yhourvar1 \
ymonadd \
ymonavg \
ymondiv \
ymoneq \
ymonge \
ymongt \
ymonle \
ymonlt \
ymonmax \
ymonmean \
ymonmin \
ymonmul \
ymonne \
ymonpctl \
ymonrange \
ymonstd \
ymonstd1 \
ymonsub \
ymonsum \
ymonvar \
ymonvar1 \
yseasadd \
yseasavg \
yseasdiv \
yseaseq \
yseasge \
yseasgt \
yseasle \
yseaslt \
yseasmax \
yseasmean \
yseasmin \
yseasmul \
yseasne \
yseaspctl \
yseasrange \
yseasstd \
yseasstd1 \
yseassub \
yseassum \
yseasvar \
yseasvar1 \
zaxisdes \
zonavg \
zonkurt \
zonmax \
zonmean \
zonmedian \
zonmin \
zonpctl \
zonrange \
zonskew \
zonstd \
zonstd1 \
zonsum \
zonvar \
zonvar1 \
zs2zl \
zs2zlx \
zsdepth \
); complete cdo 'c/-/$cdoCmpl/' 'n/*/f/'
